
/*
 * $Id: Connection.cc,v 1.39 1999/06/02 06:26:31 rousskov Exp $
 *
 *
 * Web Polygraph  http://www.ircache.net/Polygraph/
 * ----------------------------------------------------------
 *
 *  Web Polygraph is copyrighted(C) 1998 by Alex Rousskov and
 *  the University of California, San Diego (UCSD), with all 
 *  rights reserved. Development is funded by the National Science
 *  Foundation (NLANR Cache grants NCR 9616602 and NCR 9521745).
 *  Please see the COPYRIGHT file for full details.
 *
 *  Web Polygraph is free software. You may copy and distribute 
 *  VERBATIM copies of this product. You may modify your copy 
 *  for your own use. You must NOT distribute modified copies.
 *
 *  If you publish any materials derived from or related to  
 *  Web Polygraph, you must give a credit reference to the
 *  National Science Foundation and the National Laboratory 
 *  for Applied Network Research.
 *
 *  This program is distributed WITHOUT ANY WARRANTY;
 *  without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#include "polygraph.h"

#include <netinet/in.h>
#include <arpa/inet.h>

#include "Clock.h"
#include "PortMgr.h"
#include "StatPing.h"
#include "StatPhaseMgr.h"
#include "Connection.h"
#include "SharedOpts.h"
#include "polyErrors.h"
#include "globals.h"

/* Connection */

void Connection::HalfPipe::reset() {
	Assert(!theReserv);
	theBuf.reset();
	theIOCnt = 0;
	isReady = false;
}

void Connection::HalfPipe::start(Select::User *u) {
	Assert(!theReserv);
	theReserv = TheFileScanner->setFD(theConn.theSock.fd(), theDir, u);
}

void Connection::HalfPipe::stop(Select::User *) {
	if (theReserv)
		TheFileScanner->clearRes(theReserv);
}


Connection::Connection(): 
	theRd(*this, theRdBuf, dirRead),
	theWr(*this, theWrBuf, dirWrite),
	theLocPort(-1), theRemPort(-1) {
	reset();
}

void Connection::reset() {
	if (theSock)
		close();
	theAddr = NetAddr();
	//theMgr = 0;
	theRd.reset();
	theWr.reset();
	isBad = isAtEof = isLastUse = false;
	theLocPort = theRemPort = -1;

	theUseLmt = -1; // unlimited use by default
	theUseCnt = 0;
	theOpenTime = theUseStart = Time();
}

int Connection::rport() const {
	if (!theSock)
		return -1;
	else
	if (theRemPort >= 0)
		return theRemPort;
	else
		return theRemPort = theSock.rport(); // cache the result
}

bool Connection::connect(const NetAddr &addr) {
	Assert(!theSock);
	if (!theSock.create()) {
		ReportError(errno);
		return false;
	}

	setSockOpt();

	theLocPort = ThePortMgr->bind(theSock);
	if (theLocPort < 0)
		return false;

	theOpenTime = TheClock;
	CallStat(ThePinger, connOpen());

	if (!theSock.connect(theAddr = addr)) {
		ReportError(errno);
		return false;
	}
	return true;
}

bool Connection::accept(Socket &s, bool &fatal) {
	Assert(!theSock);
	sockaddr_in addr;
	int addr_len = sizeof(addr);
	theSock = s.accept((struct sockaddr*)&addr, &addr_len);
	if (!theSock) {
		fatal = errno != EWOULDBLOCK;
		if (fatal)
			ReportError(errno);
		return false;
	}

	setSockOpt();
	theAddr.addr(addr.sin_addr);
	theAddr.port(ntohs(addr.sin_port));
	theLocPort = -1; // unknown, do not care;
	theOpenTime = TheClock;
	CallStat(ThePinger, connOpen());

	return true;
}

Size Connection::read() {
	if (!theRdBuf.spaceSize()) {
		theRd.isReady = true;
		return 0;
	}
	theRd.isReady = false;

	// XXX: make first IO check optional
	const Size sz =
		(theRd.theIOCnt++ + theWr.theIOCnt) || !theSock.errNo() ?
		theSock.read(theRdBuf.space(), theRdBuf.spaceSize()) : (Size)-1;
	
	//cerr << here << "read " << sz << " : {" << theRdBuf.space() << "}" << endl;

	if (sz > 0)
		theRdBuf.appended(sz);
	else
	if (sz == 0)
		isAtEof = true;
	else
	if (sz < 0) {
		ReportError(errno);
		isBad = true;
	}
	CallStat(ThePinger, read(sz));
	return sz;
}

Size Connection::write() {
	if (!theWrBuf.contSize()) {
		theWr.isReady = true;
		return 0;
	}

	theWr.isReady = false;

	// check socket status during the first IO
	// XXX: make this check optional!
	const Size sz =
		(theRd.theIOCnt + theWr.theIOCnt++) || !theSock.errNo() ?
		theSock.write(theWrBuf.content(), theWrBuf.contSize()) : (Size)-1;
	
	//cerr << here << "wrote " << sz << " : {" << theWrBuf.content() << "}" << endl;

	if (sz > 0)
		theWrBuf.consumed(sz);
	else
	if (sz < 0) {
		ReportError(errno);
		isBad = true;
	}

	CallStat(ThePinger, wrote(sz));
	return sz;
}

bool Connection::close() {
	if (theLocPort >= 0) {
		ThePortMgr->release(theLocPort, true);
		theLocPort = -1;
	}
	if (theSock) {
		theRd.stop(0);
		theWr.stop(0);
		if (!isBad)
			CallStat(TheSoOftenPinger, connLifeTime(TheClock-theOpenTime, useCnt()));
		CallStat(ThePinger, connClose());
		if (!theSock.close()) {
			ReportError(errno);
			return false;
		}
	}
	return true;
}

// this has to be done for accepted sockets as well
// FreeBSD accept(2) man page misleads
void Connection::setSockOpt() {
	theSock.blocking(false);
	if (TheOpts.theLingerTimeout >= 0)
		theSock.linger(TheOpts.theLingerTimeout.sec());
	if (!TheOpts.doUseNagle)
		theSock.tcpDelay(false);
}

bool Connection::reusable() const {
	if (isAtEof || isLastUse || isBad)
		return false;
	return theUseLmt < 0 || theUseCnt < theUseLmt;
}

// prepare for next use if possible
// note: the notion of ``use'' is caller defined
void Connection::use() {
	Assert(reusable());
	// first user takes the hit of the opening delay
	theUseStart = theUseCnt++ ? (Time)TheClock : theOpenTime;
}
