#include "polygraph.h"

/*
 * $Id: ExpPortMgr.cc,v 1.4.2.1 1999/06/10 19:37:10 rousskov Exp $
 *
 *
 * Web Polygraph  http://www.ircache.net/Polygraph/
 * ----------------------------------------------------------
 *
 *  Web Polygraph is copyrighted(C) 1998 by Alex Rousskov and
 *  the University of California, San Diego (UCSD), with all 
 *  rights reserved. Development is funded by the National Science
 *  Foundation (NLANR Cache grants NCR 9616602 and NCR 9521745).
 *  Please see the COPYRIGHT file for full details.
 *
 *  Web Polygraph is free software. You may copy and distribute 
 *  VERBATIM copies of this product. You may modify your copy 
 *  for your own use. You must NOT distribute modified copies.
 *
 *  If you publish any materials derived from or related to  
 *  Web Polygraph, you must give a credit reference to the
 *  National Science Foundation and the National Laboratory 
 *  for Applied Network Research.
 *
 *  This program is distributed WITHOUT ANY WARRANTY;
 *  without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#include "Socket.h"
#include "ExpPortMgr.h"
#include "Rnd.h"
#include "LogComment.h"
#include "SharedOpts.h"
#include "polyErrors.h"

ExpPortMgr::ExpPortMgr(): thePortMin(-1), thePortMax(-1) {
}

ExpPortMgr::ExpPortMgr(int aPortMin, int aPortMax): thePortMin(-1), thePortMax(-1) {
	configure(aPortMin, aPortMax);
}

void ExpPortMgr::configure(int aPortMin, int aPortMax) {
	Assert(!theValidPorts.count());

	thePortMin = aPortMin;
	thePortMax = aPortMax;

	const int count = thePortMax - thePortMin + 1;
	Assert(count >= 0);
	theValidPorts.resize(count);
	theVoidPorts.resize(count);

	// push free ports, weed out used ones
	for (int p = thePortMin; p <= thePortMax; ++p)
		freePort(p, !usedPort(p));

	Assert(theValidPorts.count() + theVoidPorts.count() == count);
	// XXX: move this to report() or similar
	Comment(5) << "port mgr scanned " << count << " ports and found " << theVoidPorts.count() << " used ones" << endc;

	// randomize the order in case there is a bad subrange or something...
	RndGen rng;
	for (int i = 0; i < theValidPorts.count(); ++i)
		theValidPorts.swap(i, rng(0,theValidPorts.count()));
	for (int i = 0; i < theVoidPorts.count(); ++i)
		theValidPorts.swap(i, rng(0,theVoidPorts.count()));
}

int ExpPortMgr::allocPort(Socket &s) {
	int port;

	port = findPort(theValidPorts, s);
	if (port < 0)
		port = findPort(theVoidPorts, s);

	if (port < 0) {
		Comment << here << "no ports left;"
			<< " bound: " << theBoundLvl.level() 
			<< " valid: " << theValidPorts.count() 
			<< " void:  " << theVoidPorts.count() 
			<< " res-d: " << theResPorts.count() 
			<< " range: [" << thePortMin << ',' << thePortMax << ']'
			<< endc;
		Assert(theValidPorts.empty());
		errno = EADDRNOTAVAIL;
	} else {
		Assert(port >= thePortMin && port <= thePortMax);
	}
	return port;
}

void ExpPortMgr::freePort(int port, bool good) {
	Assert(port >= thePortMin && port <= thePortMax);
	if (good)
		theValidPorts.enqueue(port);
	else
		theVoidPorts.enqueue(port);
}

int ExpPortMgr::findPort(Ring<int> &ports, Socket &s) {
	// note: if ports == theVoidPorts, ports may "grow" inside the loop
	for (int i = ports.count(); i > 0; --i) {
		const int port = ports.dequeue();
		Assert(port >= thePortMin && port <= thePortMax);
		if (theResPorts.isSet(port)) {
			theVoidPorts.enqueue(port);
		} else
		if (!s.bind(port)) {
			const int e = errno;
			NoteError(e);
			if (ExplainError(e))
				Comment << Here << "PortMgr::findPort failed to bind to " << port << ": (" << e << ") " << strerror(e) << endc;
			theVoidPorts.enqueue(port);
		} else
			return port;
	}
	return -1;
}

bool ExpPortMgr::usedPort(int port) const {
	Socket s;
	if (s.create()) {
		bool used = !s.bind(port);
		s.close();
		return used;
	}
	return false;
}
