
/*
 * $Id: histograms.h,v 1.2 1999/06/02 06:26:40 rousskov Exp $
 *
 *
 * Web Polygraph  http://www.ircache.net/Polygraph/
 * ----------------------------------------------------------
 *
 *  Web Polygraph is copyrighted(C) 1998 by Alex Rousskov and
 *  the University of California, San Diego (UCSD), with all 
 *  rights reserved. Development is funded by the National Science
 *  Foundation (NLANR Cache grants NCR 9616602 and NCR 9521745).
 *  Please see the COPYRIGHT file for full details.
 *
 *  Web Polygraph is free software. You may copy and distribute 
 *  VERBATIM copies of this product. You may modify your copy 
 *  for your own use. You must NOT distribute modified copies.
 *
 *  If you publish any materials derived from or related to  
 *  Web Polygraph, you must give a credit reference to the
 *  National Science Foundation and the National Laboratory 
 *  for Applied Network Research.
 *
 *  This program is distributed WITHOUT ANY WARRANTY;
 *  without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#ifndef HISTOGRAMS_H
#define HISTOGRAMS_H

#include "Histogram.h"
#include "gadgets.h"

class ILog;
class OLog;

// bunch of histograms


// linear histogram (currently only 1:1 mapping is supported, no scaling)
class LineHist: public Histogram {
	public:
		LineHist() {}
		LineHist(Val aMin, Val aMax): Histogram(aMin, aMax, aMax-aMin) {}

	protected:
		virtual const char *type() const { return "line"; }
		virtual int val2Bin(Val v) const { return v; }
		virtual Val bin2Val(int b) const { return b; }
};


// log2-based histogram
// fast (no fp math!) and takes about 25KB for [0-MAX_INT) range (default)
// suitable for time and size measurements
// see the definition for details
class Log2Hist: public Histogram {
	public:
		Log2Hist();
		Log2Hist(Val aMin, Val aMax);

	protected:
		virtual const char *type() const { return "log2"; }
		virtual int val2Bin(Val v) const;
		virtual Val bin2Val(int b) const;
};


// Time histogram (internal representation uses msec)
class TimeHist: public Log2Hist {
	public:
		TimeHist();

		void record(const Time &tm) { record(Min(tm, theTmMax).msec()); }

	protected:
		void record(Val v) { Histogram::record(v); }

	protected:
		Time theTmMax;
};

typedef Log2Hist SizeHist;

// a set of histograms, actually
// ignores negative and huge sizes
class TimeVsSizeHist {
	public:
		TimeVsSizeHist();
		~TimeVsSizeHist();

		int lastGrpId() const { return szGroup(theMaxSize); }
		const TimeHist &groupHist(int g) const { return *theHists[g]; }
		int groupMaxSize(int g) const;

		void record(Size sz, const Time &tm) { if (sz >= 0 && sz <= theMaxSize) theHists[szGroup(sz)]->record(tm); }

		void store(OLog &log) const;
		void load(ILog &log);

		void add(const TimeVsSizeHist &h);

		void reset();

	protected:
		int szGroup(Size sz) const;
		
	protected:
		Array<TimeHist*> theHists;
		Size theMaxSize;
};

// a set of histograms, actually
// scales use cnt using base-2 log
class TimeVsUseHist {
	public:
		TimeVsUseHist();
		~TimeVsUseHist();

		int lastGrpId() const { return useGroup(theMaxUse); }
		const TimeHist &groupHist(int g) const { return *theHists[g]; }
		int groupMaxUse(int g) const;

		void record(const Time &tm, int useCnt) { if (useCnt > 0 && useCnt <= theMaxUse) theHists[useGroup(useCnt)]->record(tm); }

		void store(OLog &log) const;
		void load(ILog &log);

		void add(const TimeVsUseHist &h);

		void reset();

	protected:
		int useGroup(int useCnt) const;
		
	protected:
		Array<TimeHist*> theHists;
		int theMaxUse;
};

// XXX: these *Vs* histograms are all the same except initial offset and
// scale function. We should have a common class for them!
// a single histogram for each class
class TimeVsClassHist {
	public:
		static int ClassId(int connUseCnt, bool chb, bool hit);

	public:
		TimeVsClassHist();
		~TimeVsClassHist();

		int lastGrpId() const { return 5; }
		const TimeHist &groupHist(int g) const { return *theHists[g]; }

		void record(const Time &tm, int clId) { if (clId >= 0 && clId <= theMaxClass) theHists[clId]->record(tm); }

		void store(OLog &log) const;
		void load(ILog &log);

		void add(const TimeVsClassHist &h);

		void reset();

	protected:
		Array<TimeHist*> theHists;
		int theMaxClass;
};

#endif
