
/*
 * $Id: polysrv.cc,v 1.8.2.3 1999/06/14 21:08:01 rousskov Exp $
 *
 *
 * Web Polygraph  http://www.ircache.net/Polygraph/
 * ----------------------------------------------------------
 *
 *  Web Polygraph is copyrighted(C) 1998 by Alex Rousskov and
 *  the University of California, San Diego (UCSD), with all 
 *  rights reserved. Development is funded by the National Science
 *  Foundation (NLANR Cache grants NCR 9616602 and NCR 9521745).
 *  Please see the COPYRIGHT file for full details.
 *
 *  Web Polygraph is free software. You may copy and distribute 
 *  VERBATIM copies of this product. You may modify your copy 
 *  for your own use. You must NOT distribute modified copies.
 *
 *  If you publish any materials derived from or related to  
 *  Web Polygraph, you must give a credit reference to the
 *  National Science Foundation and the National Laboratory 
 *  for Applied Network Research.
 *
 *  This program is distributed WITHOUT ANY WARRANTY;
 *  without even the implied warranty of MERCHANTABILITY or
 *  FITNESS FOR A PARTICULAR PURPOSE.
 *
 */

#include "polygraph.h"

#include <new.h>

// IRIX demands _BSD_SIGNALS to be defined for signal()
// catching function declaration to have "int" as the first argument
#ifndef _BSD_SIGNALS
#define _BSD_SIGNALS
#endif
#include <signal.h>

#include <iostream.h>
#include <fstream.h>
#include <strstream.h>
#include <iomanip.h>

#include "Clock.h"
#include "Server.h"
#include "ListenQueue.h"
#include "StatPhaseMgr.h"
#include "Rnd.h"
#include "RndPermut.h"
#include "OLog.h"
#include "LogComment.h"
#include "SharedOpts.h"
#include "SrvOpts.h"

#include "gadgets.h"
#include "polyErrors.h"
#include "logTags.h"
#include "shutdown.h"
#include "globals.h"

static ListenQueue TheLQueue;

// runs one select loop cycle
static
int step(bool &idle) {
	// default: use alarm timeout
	idle = false;
	Time tout = TheAlarmClock.timeLeft();
	Time *tptr = &tout;
	if (TheAlarmClock.clear()) { // no alarms
		if (TheSrvOpts.theTimeout >= 0) { // use global timeout (if any)
			tout = TheSrvOpts.theTimeout;
			idle = true;
		} else
			tptr = 0; // no global timeout
	}

	const int readyCount = TheFileScanner->scan(tptr);
	//cerr << here << "ready files: " << readyCount << endl;
	CallStat(ThePinger, selected());

	if (readyCount < 0 && errno != EINTR)
		FatalError(errno);

	return readyCount;
}

static
void configureLogs(int prec) {
	if (TheOpts.theConsFileName && strcmp(TheOpts.theConsFileName, "-"))
		cout = cerr = clog = *new ofstream(TheOpts.theConsFileName);

	// set log file and log size
	Size logSz = TheOpts.theLogSize;
	if (TheOpts.theLogFileName) {
		TheLog().stream(TheOpts.theLogFileName, new ofstream(TheOpts.theLogFileName));
		// set default log size if needed
		if (logSz < 0)
			logSz = Size::MB(10);
	} else {
		// virtually disable logging (will use ctor default)
		if (logSz < 0)
			logSz = Size::MB(0);
	}
	Assert(logSz >= 0);
	TheLog().resize(Max(TheLog().capacity(), logSz)); // we cannot shrink log size
	TheOpts.theLogSize.set(TheLog().capacity());

	Comment.TheEchoLevel = TheOpts.theVerbLevel;

	configureStream(cout, prec);
	configureStream(cerr, prec);
	configureStream(clog, prec);
}

// set some general stuff and
// propogate cmd line options to corresponding objects
static
void configure() {

	configureLogs(2);

	atexit(&Shutdown);

	ErrorMgr::Configure(new PolyErrMgr);

	// ignore some signals
	signal(SIGPIPE, SIG_IGN);
	signal(SIGHUP, SIG_IGN);

	// quit nicely on some signals
	signal(SIGINT, &ShutdownSignal);

	// set resource limits
	const int rlimit = SetMaxFD(GetMaxFD());
	if (rlimit > FD_SETSIZE)
		cerr << here << "warning: getrlimit(2) system call reports " << rlimit << " file descriptors while FD_SETSIZE #defines only " << FD_SETSIZE << "; using lower value." << endl;
	const int fdLimit = Min(FD_SETSIZE, rlimit);
	if (fdLimit <= 0)
		FatalError(errZeroFDLimit);
	TheFileScanner = TheOpts.theFileScanner.val();
	TheFileScanner->configure(fdLimit);
	TheFileScanner->ticker(&TheLQueue);

	Socket::TheMaxLevel = Max(0, fdLimit - 10 - fdLimit/33);
	if (TheOpts.theFDLimit >= 0 && TheOpts.theFDLimit < Socket::TheMaxLevel)
		Socket::TheMaxLevel = TheOpts.theFDLimit;
	TheOpts.theFDLimit.set(Socket::TheMaxLevel);

	// set random seeds
	RndPermutSeed(TheOpts.theRngSeed);
	RndGen::TheDefSeed = RndPermut(TheOpts.theRngSeed);

	// configure object life cycle model
	TheSrvOpts.theObjLifeCycle.bday(TheSrvOpts.theObjBday.condDistr());
	TheSrvOpts.theObjLifeCycle.modVar(TheSrvOpts.theModTimeVar);
	TheSrvOpts.theObjLifeCycle.modTime(TheSrvOpts.theModTime.condDistr(), TheSrvOpts.theShowModRatio);

	// XXX: kludge, fix me!
	if (TheOpts.theMonitor) {
		extern NetAddr TheMonitor;
		extern String TheRunLabel;

		TheMonitor = TheOpts.theMonitor;
		TheRunLabel = TheOpts.theRunLabel;
	}

	TheStatPhaseMgr.goal(TheOpts.theGoal);
}

int main(int argc, char *argv[]) {

	CmdLine cmd(Array<OptGrp*>() << &TheOpts << &TheSrvOpts);
	if (!cmd.parse(argc, argv) || !TheOpts.validate() || !TheSrvOpts.validate())
		return -1;

	Clock::Update(false);

	configure();

	cmd.report();

	Comment(2) << "FDs: "
		<< GetCurFD() << " out of " << GetMaxFD()
		<< " FDs can be used; safeguard limit: " << Socket::TheMaxLevel 
		<< endc;

	const UniqId runId = UniqId::Create();
	Comment(1) << "run-id: " << runId << " pid: " << getpid() << endc;
	TheLog() << bege(lgRunId) << runId << ende;

	Clock::Update();
	Comment << "waiting for requests; goal: " << TheOpts.theGoal << endc;

	ThePinger.start();
	TheSoOftenPinger.start();

	Server TheServer;
	TheServer.configure(TheSrvOpts.thePort, TheOpts.theXactThinkTime.condDistr());
	TheServer.listenQueue(&TheLQueue);
	TheServer.start();

	while (!ShutdownNow()) {
		Clock::Update();
		ThePinger.checkpoint();
		TheSoOftenPinger.checkpoint();
		bool idle;
		if (!step(idle) && idle) {
			Comment << "was idle for at least " << TheSrvOpts.theTimeout << endc;
			ShutdownReason("incativity timeout");
		}
	}

	return 0;
}
